/*
 *  regex
 *  Partial implementation of regexp's.
 *
 *  Copyright (C) 2000 Jens Laas jens.laas@data.slu.se
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

// #define REGEX_MAIN
// #define TESTING

/*
       In  basic  regular expressions the metacharacters ?, +, {,
       |, (, and ) lose their special meaning;  instead  use  the
       backslashed versions \?, \+, \{, \|, \(, and \).

       In  egrep  the  metacharacter { loses its special meaning;
       instead use \{.
*/

#include "regex.h"

const int debug = 0;
int regex_mode = MODE_GREP;

void print_expart(expart *, int);

enum {
  EX_NONE,
  EX_CHAR,
  EX_ANY,
  EX_MAP,
  EX_SELECT,
  EX_SELPART,
  EX_BOL,
  EX_EOL };

enum {
  ENUM_1,
  ENUM_01,
  ENUM_0PLUS,
  ENUM_1PLUS,
  ENUM_SUB };

typedef unsigned char uchar;

char   *extract_selector(char *string);
char  **split_selector(char *s);
expart *build_map(char **string);
expart *build_selector(char **string);
char   *query_regex(expart *p, char *s);
expart *build_regex( char *string );
void    finalize(expart *p, struct regex *r);


struct regex *regex( char *string )
{
  struct regex *r;
  char *s;
  
  /* add default .* infront of expression */
  s = malloc( strlen( string) + 3);
  strcpy(s, ".*");
  strcat(s, string);
  
  r = malloc(sizeof( struct regex ) );
  r->start = 0;
  r->head = build_regex( s );
  finalize( r->head, r);
  free( s );
  if(debug) {
    printf("built regex\n");
    print_expart(r->head, 0);
  }
  return r;
}

int query( struct regex *regex, char *string )
{
  static int count = 0;
  int m;
  
  count++;
  
  if(!regex)
    return 0;
  regex->start = string;
  if(debug)
    {
      printf("start query %d against: %s\n", count, string);
      print_expart(regex->head, 0);
    }
  
  if(query_regex( regex->head, string ) == 0)
    m = 0;
  else
    m = 1;
  regex->start = 0;
  if(debug) printf("end query: %s\n", m?"MATCH":"--");
  return m;
}

void finalize(expart *p, struct regex *r)
{
  p->regex = r;
  if(p->subparts)
    finalize(p->subparts, r);
  if(p->next)
    finalize(p->next, r);
}

expart *new_expart( )
{
  expart *p;
  p = malloc( sizeof( expart ) );
  if(!p)
    exit(2);
  
  memset(p, 0, sizeof( expart ));
  
  return p;
}

int enumeration(char **s)
{
  if(regex_mode == MODE_EGREP)
    switch( **s )
      {
      case '?':
	(*s)++;
	return ENUM_01;
      case '+':
	(*s)++;
	return ENUM_1PLUS;
      }
  if(**s == '*')
    {
      (*s)++;
      return ENUM_0PLUS;
    }
  return ENUM_1;
}

void set_enumeration( expart *p, int e )
{
  p->enumeration = e;
  if(p->type == EX_SELECT)
    {
      for(p = p->subparts; p; p = p->next)
	p->enumeration = ENUM_SUB;
      //p->enumeration = e;
    }
  return;
}

expart *build_regex( char *string )
{
  expart *p, *base = 0, *last = 0;
  char *s = string;
  
  while(*s)
    {
      switch(*s)
	{
	case '(':
	  if(regex_mode == MODE_EGREP) {
	    p = build_selector(&s);
	    set_enumeration(p, enumeration(&s));
	    break;
	  }
	case '[':
	  p = build_map(&s);
	  set_enumeration(p, enumeration(&s));
	  break;
	case '.':
	  s++;
	  p = new_expart();
	  p->type = EX_ANY;
	  set_enumeration(p, enumeration(&s));
	  break;
	case '^':
	  s++;
	  p = new_expart();
	  p->type = EX_BOL;
	  p->enumeration = ENUM_1;
	  break;
	case '$':
	  s++;
	  p = new_expart();
	  p->type = EX_EOL;
	  p->enumeration = ENUM_1;
	  break;
	case '\\':
	  s++;
	  /* fall thru to ordinary char */
	default:
	  /* ordinary char */
	  p = new_expart();
	  p->type = EX_CHAR;
	  p->c = *s++;
	  set_enumeration(p, enumeration(&s));
	  break;
	}
      
      if(!base)
	base = p;
      else
	last->next = p;
      last = p;
    }
  
  return base;
}


expart *build_selector(char **string)
{
  expart *p, *base, *last = 0;
  char *s, **vec;

  base = new_expart();
  base->type = EX_SELECT;
  
  /* split selector into substrings */
  s = extract_selector(*string);
  /* move the "stream" */
  (*string) += strlen(s) + 2;
  
  vec = split_selector(s);
  
  /* for each substring build a sub-regex */
  while(*vec)
    {
      p = new_expart();
      p->type = EX_SELPART;
      p->sub_string = *vec;
      p->subparts = build_regex(*vec++);
      if(!base->subparts)
	base->subparts = p;
      else
	last->next = p;
      base->nr_subparts++;
      last = p;
    }
  return base;
}

expart *build_map(char **string)
{
  expart *p;
  uchar *s = (*string)+1;
  uchar prev = 0;
  int sign = 1;

  if(debug)
    printf("build map %s\n", s);
  
  p = new_expart();
  p->type = EX_MAP;
  p->map = (char *) malloc(256);
  if(!p->map)
    exit(2);
  
  /* first char cases */
 again:
  switch(*s)
    {
    case '-':
    case ']':
      p->map[(unsigned int)*s] = sign;
      prev = *s;
      s++;
      break;
    case '^':
      sign = 0;
      s++;
      memset(p->map, 1, 256);
      goto again;
    }

  p->sign = sign;
  
  while(*s)
    {
      switch(*s)
	{  
	case ']':
	  *string = s+1;
	  return p;
	case '-':
	  s++;
	  while(prev <= *s)
	    p->map[prev++] = sign;
	  break;
	default:
	  p->map[*s] = sign;
	  prev = *s;
	}
      s++;
    }
  return p;
}


/*
 * Extracts the selector (xxx) from string
 */
char *extract_selector(char *string)
{
  /* count () and ignore in [] */
  int paren = 0;
  int map = 0;
  char *s = strdup(string);
  char *sel = s;

  if(debug)
    printf("extract sel %s\n", string);
  
  while(*s)
    {
      switch(*s++)
	{
	case '(':
	  if(!map)
	    paren++;
	  break;
	case ')':
	  if(!map)
	    {
	      paren--;
	      if(paren == 0)
		{
		  *(s-1) = 0;
		  return sel+1;
		}
	    }
	  break;
	case '[':
	  if(!map)
	    map = 1;
	  break;
    	case ']':
	  map = 0;
	  break;
	case '\\':
	  if(!map)
	    s++;
	  break;
	}
    }
  return 0;
}

char **split_selector(char *string)
{
  int paren = 0;
  int map = 0;
  char **vec, **v;
  char *start, *s, *s2;
  int n = 2;
  
  if(debug)
    printf("split %s\n", string);
  
  s = strdup(string);
  if(!s)
    exit(2);
  start = s;
  
  s2 = s;
  while( (s2 = strchr(s2, '|')) != 0)
    {
      s2++;
      n++;
    }
  
  /* detect '|' ignore in () and [] */
  vec = malloc( n * sizeof(char *));
  if(!vec)
    exit(2);
  memset(vec, 0, n * sizeof(char *));
  
  v = vec;
  
  while(*s)
    {
      switch(*s++)
	{
	case '(':
	  if(!map)
	    paren++;
	  break;
	case ')':
	  if(!map)
	    paren--;
	  break;
	case '[':
	  if(!map)
	    map = 1;
	  break;
    	case ']':
	  map = 0;
	  break;
	case '|':
	  if(!(paren || map))
	    {
	      *v++ = start;
	      *(s-1) = 0;
	      start = s;
	    }
	  break;
	case '\\':
	  s++;
	  break;
	}
    }
  
  *v++ = start;
  *v = 0;
  return vec;
}


/*
 * query a single expart for match
 * Return a vector of match(es).
 */
char **query_single(expart *p, char *s)
{
  char **vec, **v, *end;
  expart *sp;

  if(debug)
    printf("nr subparts: %d aginst: %s\n", p->nr_subparts, s);
  
  vec = malloc( (p->nr_subparts+2) * sizeof(char *));
  
  if(!vec)
    exit(2);
  v = vec;
  
  switch(p->type)
    {
    case EX_CHAR:
      if(debug)
	printf("EX_CHAR [%c] <%s>\n", p->c, s);
      if(p->c == *s)
	*v++ = s+1;
      break;
      
    case EX_ANY:
      if(debug)
	printf("EX_ANY <%s>\n", s);
      if(*s)
	*v++ = s+1;
      break;
      
    case EX_MAP:
      if(debug)
	printf("EX_MAP <%s>\n", s);
      if(p->map[(unsigned int)*s])
	*v++ = s+1;
      break;
      
    case EX_SELECT:
      if(debug)
	printf("EX_SELECT <%s>\n", s);
      /* ask each subpart for success */
      for(sp = p->subparts; sp; sp = sp->next)
	{
	  if(debug)
	    printf("-- Asking subpart\n");
	  end = query_regex(sp, s);
	  if(end) {
	    *v++ = end;
	    if(debug)
	      printf("Subpart success!\n");
	  }
	  else
	    {
	      if(debug)
		printf("Subpart failed!\n");
	    }
	}
      break;
      
    case EX_SELPART:
      if(debug)
	printf("EX_SELPART [%s] <%s>\n", p->sub_string, s);
      *v++ = query_regex(p->subparts, s);
      break;
      
    case EX_BOL:
      if(debug)
	printf("EX_BOL <%s>\n", s);
      if(s == p->regex->start)
	*v++ = s;
      break;
      
    case EX_EOL:
      if(debug)
	printf("EX_EOL <%s>\n", s);
      if(*s == 0)
	*v++ = s;
      break;
    default:
      if(debug) printf("UNKNOWN TYPE\n");
    }
  if(debug)
    printf("passed switch\n");
  *v = 0;
  if(debug)
    printf("hits %d\n", v-vec);
  return vec;
}


char *query_regex(expart *p, char *s)
{
  char *end, **vec, **v;

  if(!p) {
    if(debug)
      printf("NULL expart\n");
    return s;
  }

  if(debug)
    printf("do query_single()\n");
  vec = query_single(p, s);
  
  if(debug)
    if(!*vec)
      printf("Match fail, 0\n");
  if(debug) printf("%s\n", *vec?*vec:"(NULL)");
  
  switch(p->enumeration)
    {
    case ENUM_SUB:
      free(vec);
      return *vec;
      
    case ENUM_1PLUS:
      if(!*vec)
	{
	  free(vec);
	  return 0;
	}
      v = vec;
      while(*v)
	{
	  if( (end = query_regex(p, *v)) ) {
	    /* can we match this one and the rest? */
	    free(vec);
	    return end;
	  }
	  
	  /* can we match the rest of the regex? */
	  if( (end = query_regex(p->next, *v)) )
	    return end;
	  v++;
	}
      break;
      
    case ENUM_0PLUS:
      if(!*vec)
	/* we failed this one, can we match the rest? */
	{
	  free(vec);
	  return query_regex(p->next, s);
	}
      v = vec;
      while(*v)
	{
	  if( (end = query_regex(p, *v)) ) {
	    free(vec);
	    return end;
	  }
	  if( (end = query_regex(p->next, *v)) ) {
	    free(vec);
	    return end;
	  }
	  v++;
	}
      free(vec);
      return query_regex(p->next, s);
      break;
      
    case ENUM_1:
      if(debug)
	printf("ENUM_1\n");
      if(!*vec)
	{
	  free(vec);
	  return 0;
	}
      v = vec;
      while(*v)
	{
	  if(debug)
	    printf("Try next\n");
	  if( (end = query_regex(p->next, *v)) ) {
	    free(vec);
	    return end;
	  }
	  v++;
	}
      break;
      
    case ENUM_01:
      if(!*vec)
	return query_regex(p->next, s);
      v = vec;
      while(*v)
	{
	  if( (end = query_regex(p->next, *v)) ) {
	    free(vec);
	    return end;
	  }
	  v++;
	}
    }
  free(vec);
  if(debug)
    printf("Enumeration Match failed\n");
  return 0;
}

#ifdef TESTING
/**************************************************************
 * Functions for testing purposes
 **************************************************************
 */
void print_vector(char **v)
{
  while(*v)
    printf("<%s>\n", *v++);
}

char *types[] = {
  "EX_NONE", "EX_CHAR", "EX_ANY", "EX_MAP", "EX_SELECT", "EX_SELPART", "EX_EOL"
};

char *enums[] = {
  "ENUM_1", "ENUM_01", "ENUM_0PLUS", "ENUM_1PLUS"
};

void print_expart(expart *p, int indent)
{
  int i;
  if(!p)
    return;
  
 more:
  for(i=0;i<indent;i++)
    printf(" ");
  
  printf("%s %s ", types[p->type], enums[p->enumeration]);
  if(p->type == EX_CHAR)
    printf("[%c] ", p->c);

  if(p->type == EX_MAP)
    {
      int i;
      printf("[");
      if(!p->sign)
	printf("^");
      for(i=0; i<256; i++)
	if(p->map[i] == p->sign)
	  printf("%c", i);
      printf("]");
    }
  
  printf("\n");
  
  if(p->type == EX_SELECT)
    print_expart(p->subparts, indent +2);
  
  if(p->type == EX_SELPART)
    print_expart(p->subparts, indent +1);
  
  if(p->next)
    {
      p = p->next;
      goto more;
    }
}
#else
void print_expart(expart *p, int i) { }
#endif
#ifdef REGEX_MAIN
int main()
{
  char b[256], s[256];

  /*
{
  char **v;
  v = split_selector("apa|groda|fisk");
  print_vector(v);
  v = split_selector("apa|(g|r|o|d|a)|fisk");
  print_vector(v);
  v = split_selector("apa|[|]|fisk");
  print_vector(v);
}  
*/

  /*
  printf("<%s>\n", extract_selector("(hej|hopp)abrakadabra"));
  printf("<%s>\n", extract_selector("(hej|(h|o|p|p))sadas"));
  printf("<%s>\n", extract_selector("(hej|[[ho(pp])asdasd"));
  */
  
  while(1)
    {
      scanf("%s", b);
      scanf("%s", s);
      printf("%s\n", query(regex(b), s) == 0 ? "false" : "true");
    }

  return 0;
}
#endif

/*

AdobedobeA

*/

/*

gcc -Wall -o regex regex.c

gcc -Wall -Os -c regex.c
*/

