/*
 *  ugrep
 *  barebones version of grep.
 *
 *  Copyright (C) 2000 Jens Laas jens.laas@data.slu.se
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * Changes:
 * v 1.0 First release
 * v 1.1 Matching for ^ (beginning of line)
 * v 1.2 Matching for $ (end of line)
 * v 1.3 -h switch added
 * v 1.4 -q -s switches added
 * v 1.5 -l -L switches added
 * v 1.6 2001-01-10 partial egrep support added.
 * v 1.7 2001-04-19 removed stdio dependence (down to 7396 bytes!)
 *   Missing: {}, [:name:], \w, \W, \b, \B, \<, \> and backref \n
 *
 */

/*
  Supported commandline options:
  
  -i  ignore case
  -C  context
  -n  line numbers
  -v  reverse match
  -V  print versionint conf_filename = 0;
  -h  suppress prefix of filenames
  -q  silent mode. Exit after first match.
  -s  no (error) messages
  -l  print files with matches
  -L  print files without matches
  
  Pattern:
  ugrep will only do basic regexp string matching.
  
*/

#define VERSION "1.7"

#define USE_REGEXP

#define BANNER "\nCopyright (C) 2000 Jens Laas\nThis is free software; see the source for copying conditions. There is NO\nwarranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\n"

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>

#include "uio.h"

#ifdef USE_REGEXP
#include "regex.h"
extern int regex_mode;
#endif

struct ugrep_config {
  int context;
  int lineno;
  int reverse;
  int ignore_case;
  int filename;
  int no_filename;
  int stop_on_first;
  int quiet;
  int no_messages;
  int files_without_match;
  int files_with_matches;
#ifdef USE_REGEXP
  struct regex *regexp;
#endif
};

struct ugrep_config conf = {
  context: 0,
  lineno: 0,
  reverse: 0,
  ignore_case: 0,
  filename: 0,
  no_filename: 0,
  quiet: 0,
  no_messages: 0,
  files_without_match: 0,
  files_with_matches: 0,
  stop_on_first: 0
};

/*
 * Convert string to uppercase
 */
#if 0
static char *lower_case = "abcdefghijklmnopqrstuvwxyz";
static char *upper_case = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
#endif
void strupper(char *s)
{
  /* char *r; */
  
  while(*s)
    {
#if 0
      if( (r = strchr(lower_case, *s)) )
	*s = upper_case[r-lower_case];
#endif
      if(*s >= 'a' && *s <= 'z')
	*s -= 32;
      s++;
    }
}

/*
 * Test file if it is a text or binary.
 * Reads first 32 bytes and checks for values under 5
 */
int istext(int fd)
{
  unsigned char b[32];
  int l;

  l = read(fd, b, 32);
  if(l>0)
    {
      while(--l)
	if(b[l] < 5)
	  return 0;
    }
  else
    return 0;
  
  lseek(fd, 0, SEEK_SET);
  return 1;
}

/*
 * reads one line from fd
 * keeps only l number of bytes of the line.
 * always reads until end of line (or file)
 */
int read_line(int fd, char *b, int l)
{
  int got = 0, i;
  
  while( (i = read(fd, &b[got], 1)) > 0)
    {
      if(b[got] == '\n')
	{
	  b[got] = 0;
	  return 1;
	}
      
      if(got < (l-1))
	got += i;
    }
  
  b[got] = 0;
  return 0;
}

/*
 * Prints a matching line to stdout.
 * All formatting of line done here.
 */
char *current_file;
void println( char *line, int lineno, int match)
{
  if(!line)
    return;

  if(conf.quiet)
    return;
  
  if(conf.lineno)
    {
      vcat( 2, conf.filename ? current_file : "",
	    conf.filename ? ":" : "");
      cat_int(lineno);
      vcat( 4, ":", (match && conf.context) ? ">>>":"",
	    line, "\n");
    }
  else
    {
      vcat(5, conf.filename ? current_file : "",
	   conf.filename ? ":" : "",
	   (match && conf.context) ? ">>>":"",
	   line, "\n");
    }
  print((char *) 1);
}

/*
 * tests one line for match against pattern
 * returns 0 on failed match
 */
int match_line(char *s, char *p)
{
#ifndef USE_REGEXP
  int r, l;
  char *p2;
#endif

  /* return 0 if failed match */

#ifdef USE_REGEXP
  if(!conf.regexp)
    conf.regexp = regex( p );
  return query( conf.regexp, s );
#endif
  
#ifndef USE_REGEXP
  l = strlen(p);
  
  if((*p == '^') && (*(p+l-1) == '$'))
    {
      /* match beginning and end: exact match */
      p2 = strdup(p);
      
      *(p2+l-1) = 0; /* remove trailing '$' */
      r = strcmp(s,p2+1);
      free(p2);
      return !(r);
    }
  
  if(*(p+l-1) == '$')
    {
      /* match end */
      
      p2 = strdup(p);
      *(p2+l-1) = 0; /* remove trailing '$' */
      
      if(strlen(s) < (l-1) )
	return 0;
      
      r = strcmp(s+strlen(s)-l+1, p2);
      free(p2);
      return !(r);
    }
  
  if(*p == '^')
    {
      /* match beginning */
      char *s2;
      if(strlen(s) < strlen(p+1) )
	return 0;
      s2 = strdup(s);
      s2[ strlen(p+1) ] = 0;
      r = strcmp(s2, p+1);
      free(s2);
      return !(r);
    }
  
  /* match anywhere */
  return (int) strstr(s, p);
#endif
}
/*
  Some lines for testing greps:
  abrakadabra int
int
*/

/*
 * Match given file against pattern
 * if filname == 0 then test against stdin
 */
int match_file( char *pattern, char *filename)
{
  int fd, n = 0, no = 0;
  char b[1024], *b2;
  char *context[2] = { 0,0 };
  int match = 0;
 
  current_file = filename;
  
  if(filename)
    fd = open(filename, O_RDONLY);
  else
    fd = 0;
  
  if(fd < 0)
    {
      if(!(conf.quiet || conf.no_messages))
	{
	  vcat(3, "File ", filename, " not found.\n");
	  print((char *)2);
	}
      return 0;
    }

  if(istext(fd))
    {
      /* loop while there are lines to read */
      while(read_line(fd, b, 1024))
	{
	  int result;
	  
	  no++;
	  
	  b2 = strdup(b); /* use b2 for case-insensitive matching */
	  if(conf.ignore_case)
	    strupper(b2);
	  
	  result = match_line(b2, pattern);
	  free(b2);
	  result = conf.reverse ? (!result) : result;

	  if( result )
	    {
	      match = 1;
	      
	      if(conf.context)
		{
		  if(context[0])
		    {
		      cat("--\n");
		      print((char *)1);
		    }
		
		  println(context[0], no-2, 0);
		  println(context[1], no-2, 0);
		  n = 2;
		}
	      
	      println(b, no, 1);
	    }
	  
	  /* 
	   * Context handling
	   * Looks messy, I know
	   */
	  if( (!result) && n)
	    {
	      n--;
	      println(b, no, 0);
	    }
	  else
	    {
	      if(context[0])
		free(context[0]);
	      context[0] = context[1];
	      if(result)
		{
		  context[1] = 0;
		  if(context[0])
		    free(context[0]);
		  context[0] = 0;
		}
	      else
		context[1] = strdup(b);
	    }
	}
    }
  else
    {
      if(fd > 0)
	{
	  if(!(conf.quiet || conf.no_messages))
	    {
	      vcat(3, "Binary file ", filename, ".\n");
	      print((char *)2);
	    }
	}
    }
  
  close(fd);
  return match;
}

int main(int argc, char  **argv)
{
  char *pattern = 0;
  int use_stdin = 1;
  int match = 0;
  char *me2, *me = *argv;

  init_io();

  for(me2 = me; *me2; me2++)
    if(*me2 == '/')
	me = me2+1;
  
  if(!strcmp(me, "egrep") || !strcmp(me, "uegrep")) {
    regex_mode = MODE_EGREP;
  }
  
  if(argc < 2)
    goto usage;  
  
  argv++;

  while(--argc)
    {
      if(**argv == '-')
	{
	  while(*++(*argv))
	    {
	      switch(**argv)
		{
		case 'C':
		  conf.context = 1;
		  break;
		case 'n':
		  conf.lineno = 1;
		  break;
		case 'v':
		  conf.reverse = 1;
		  break;
		case 'i':
		  conf.ignore_case = 1;
		  break;
		case 'q':
		  conf.quiet = 1;
		  conf.stop_on_first = 1;
		  break;
		case 's':
		  conf.no_messages = 1;
		  break;
		case 'h':
		  conf.no_filename = 1;
		  break;
		case 'L':
		  conf.files_without_match = 1;
		  conf.quiet = 1;
		  break;
		case 'l':
		  conf.files_with_matches = 1;
		  conf.quiet = 1;
		  break;
		case 'V':
		  goto usage;
		case '-':
		  goto usage;
		}
	    }
	  argv++;
	  continue;
	}
      
      if(pattern)
	{
	  if((argc > 1) && (conf.no_filename == 0))
	    conf.filename = 1;
	  match |= match_file(pattern, *argv++); 
	  use_stdin = 0;
	  
	  if( match && conf.files_with_matches )
	    {
	      vcat(2, current_file, "\n");
	      print((char *)1);
	    }
	  if( (!match) && conf.files_without_match )
	    {
	      vcat(2, current_file, "\n");
	      print((char *)1);
	    }
	  
	  if(match && conf.stop_on_first)
	    break;
	}
      else
	{
	  pattern = *argv++;
	  if(conf.ignore_case)
	    strupper(pattern);
	}
    }
  
  if(use_stdin)
    {
      if(pattern)
	match = match_file(pattern, 0); 
      else
	goto usage;
    }
  
  if(match)
    return 0;
  return 1;
  
 usage:
  vcat(6 , "ugrep v", VERSION, "\nUsage: ", me,
       " [-iCnvVh] pattern [file ...]\n", BANNER);
  print((char *)2);
  return 1;
}

/*
  Dynamically linked:

  gcc -Wall -Os -s -o ugrep ugrep.c

  Statically linked: (to dietlibc)
  gcc -Wall -Os -c ugrep.c
  gcc -nostdlib -s -o ugrep ugrep.o ~/misc/dietlibc/start.o ~/misc/dietlibc/dietlibc.a 

*/








